import matplotlib.pyplot as plt

from astropy.wcs import WCS
import astropy.units as u
from astropy.io import fits
from astropy.visualization.wcsaxes import add_beam, add_scalebar

plt.ion()

import numpy as np


from velocity_tools import vgrad

plt.rcParams.update({"text.usetex": True,
                     "font.family": "serif",
                     'xtick.direction': 'in',
                     'ytick.direction': 'in',
                     'xtick.major.size': 6,
                     'ytick.major.size': 6,
                     'xtick.minor.size': 3,
                     'ytick.minor.size': 3})

def plot_setup_wcsaxes(ax, header, label_col='black'):
    scalebar_length = 100*u.au
    scalebar_text="100 au"
    tick_minor_x = 4
    tick_minor_y = 6
    # ticks
    ra_ax = ax.coords[0] # lon
    dec_ax = ax.coords[1] # lat
    ra_ax.set_major_formatter('hh:mm:ss')
    dec_ax.set_major_formatter('dd:mm:ss')
    ra_ax.display_minor_ticks(True)
    dec_ax.display_minor_ticks(True)
    ra_ax.set_minor_frequency(tick_minor_x)
    dec_ax.set_minor_frequency(tick_minor_y)
    ax.tick_params(which='major', length=6)
    ax.tick_params(which='minor', length=3)
    ax.autoscale(enable=False)
    # Add beamsize
    add_beam(ax, header=header, frame=False, pad=0.5, 
             color=label_col, corner='top right')
    # # Scalebar
    length = (scalebar_length / my_distance).to(
        u.deg, u.dimensionless_angles()
        )
    add_scalebar(ax, length, label=scalebar_text, color=label_col, 
        corner='bottom right')
    #
    ax.set_xlabel(r'Right Ascension (J2000)')
    ax.set_ylabel(r'Declination (J2000)')
    return #fig_i



my_figsize = (6.5, 5)
my_distance = 10.0*u.pc

# 111 x 111 image
# pixel of 1 arcsec
# delta x is negative, R.A.
n_pix = 111
x_t = -1* np.arange(n_pix) * u.arcsec#.to(u.deg)
y_t = np.arange(n_pix) * (u.arcsec)#.to(u.deg)
xv, yv = np.meshgrid(x_t, y_t, indexing='xy')
gx = -2.0 * u.km / u.s / u.deg
gy = -2.0 * u.km / u.s / u.deg
v0 = 8.5 * u.km / u.s
ev0 = 0.1 * u.km / u.s
v_t = (xv * gx + yv * gy).to(u.km / u.s) + v0 # km/s
ev_t = np.empty(v_t.shape) * u.km / u.s
ev_t[:] = ev0 # km/s

hdu = fits.PrimaryHDU(v_t.value)
#
hdu.header['CDELT1'] = (-1 * u.arcsec.to(u.deg), 'deg')
hdu.header['CDELT2'] = (1 * u.arcsec.to(u.deg), 'deg')
hdu.header['CRPIX1'] = (len(x_t.value) * 0.5, 'Pixel reference')
hdu.header['CRPIX2'] = (len(y_t.value) * 0.5, 'Pixel reference')
hdu.header['CRVAL1'] = (15*3.5, 'deg')
hdu.header['CRVAL2'] = (31 + 1./3., 'deg')
hdu.header['CTYPE1'] = 'RA---TAN'
hdu.header['CTYPE2'] = 'DEC--TAN'
#
hdu.header['BUNIT'] = ('km / s', 'Vlsr')
hdu.header['BMAJ'] = 2 * hdu.header['CDELT2']
hdu.header['BMIN'] = hdu.header['BMAJ']
hdu.header['BPA'] = 0.0

file_vc = 'test_vel.fits'
file_evc = 'test_vel_err.fits'
hdu.writeto(file_vc, overwrite=True)
hdu.data = ev_t.value
hdu.writeto(file_evc, overwrite=True)

result_all = vgrad.vfit(xv, yv, v_t, ev_t, distance=10.0*u.pc)
# 1 parsec = 1 * u.pc.to(u.au) 
# = (1 * u.pc.to(u.au) / 10. * u.arcsec).to(u.deg)
conv_grad = (1 * u.pc.to(u.au) / 10. * u.arcsec).to(u.deg) / u.pc
input_grad = conv_grad * np.sqrt(gx**2 + gy**2)
input_PA = np.rad2deg(np.arctan2(gx.value, gy.value))

print('VGrad input = {0:4.2f}    VGrad output = {1:4.2f}'.format(input_grad, 
                        result_all['grad']))

result = vgrad.vfit_image(file_vc, file_evc, distance=10.0*u.pc, 
                          n_oversample=7, width=3*u.arcsec)
print('VGrad input = {0:4.2f}    VGrad output = {1:4.2f}'.format(input_grad, 
                        result['header']['VG']))

print('Input PA = {0:5.2f}'.format(input_PA))
#
#

# coordinate system for all images
wcs = WCS(hdu.header)

plt.close('all')
#
# Velocity map
#
fig1 = plt.figure(figsize=my_figsize)
ax1 = plt.subplot(projection=wcs)
im1 = ax1.imshow(fits.getdata(file_vc), cmap='RdYlBu_r', interpolation='none')
plot_setup_wcsaxes(ax1, hdu.header, label_col='black')
ax1.text(0.1, 0.9, 'Vlsr', transform=ax1.transAxes)
fig1.colorbar(im1, ax=ax1)


# This requires to fix display for bad/edge data points
cmap = plt.get_cmap(name='inferno').copy()
cmap.set_bad(color='0.85')

#
# Gradient amplitude map
#
fig2 = plt.figure(figsize=my_figsize)
ax2 = plt.subplot(projection=wcs)
im2 = ax2.imshow(result['grad'], cmap=cmap, interpolation='none',
    vmin=0.99 * input_grad.value, vmax=1.01 * input_grad.value)
plot_setup_wcsaxes(ax2, hdu.header, label_col='black')
ax2.text(0.1, 0.9, 'Gradient Vlsr', transform=ax2.transAxes)
fig2.colorbar(im2, ax=ax2)


fig3 = plt.figure(figsize=my_figsize)
ax3 = plt.subplot(projection=wcs)
im3 = ax3.imshow(result['posang'], cmap=cmap, interpolation='none',
    vmin=input_PA - 5, vmax=input_PA + 5)
plot_setup_wcsaxes(ax2, hdu.header, label_col='black')
ax3.text(0.1, 0.9, 'Gradient PA', transform=ax3.transAxes)
fig3.colorbar(im3, ax=ax3)

#
# Add quivers to velocity map, while undersampling the gradient results
ns = 10
x_map = np.arange(0, hdu.header['NAXIS1'])
y_map = np.arange(0, hdu.header['NAXIS2'])
xv_map, yv_map = np.meshgrid(x_map, y_map)
ra_map, dec_map = wcs.wcs_pix2world(xv_map, yv_map, 0)
#
gd_grad = np.isfinite(result['grad']) & (xv_map % ns == 0) & (yv_map % ns == 0)
ra_grad = ra_map[gd_grad]
dec_grad = dec_map[gd_grad]
grad_abs = result['grad'][gd_grad]
grad_pa = result['posang'][gd_grad]
#
grad_x = grad_abs * np.cos(np.deg2rad(grad_pa + 90))
grad_y = grad_abs * np.sin(np.deg2rad(grad_pa + 90))
"""
I don't know how to plot the quivers in sky unit coordinates
ax1.quiver(ra_grad*u.deg, dec_grad*u.deg, grad_x*u.deg, grad_y*u.deg, scale=10, 
    transform=ax1.get_transform('world'))
"""
quiv = ax1.quiver(xv_map[gd_grad], yv_map[gd_grad], grad_x, grad_y, scale=5, 
    pivot='mid', width=0.003, scale_units='xy', angles='xy', 
    headlength=3.5, headaxislength=3.0)
ax1.quiverkey(quiv, X=0.5*hdu.header['NAXIS1'], Y=1.03*hdu.header['NAXIS2'], 
    U=10, label=r'10 km s$^{-1}$ pc$^{-1}$', coordinates='data')



import os
import numpy as np

from astropy.io import fits
from astropy import units as u
from astropy.io import fits
import velocity_tools.vgrad
import pytest

my_distance = 10.0 * u.pc
v0 = 8.5 * u.km / u.s
g0 = 2.0 * u.km / u.s / u.deg
gx = -g0
gy = g0


@pytest.fixture
def velocity_file(tmp_path) -> list:
    """
    make gradient and then
    """
    half_size = 20
    n_pix = 2 * half_size + 1

    x_t = -1 * np.arange(n_pix) * u.arcsec
    y_t = np.arange(n_pix) * (u.arcsec)
    xv, yv = np.meshgrid(x_t, y_t, indexing="xy")
    ev0 = 0.1 * u.km / u.s
    v_t = (xv * gx + yv * gy).to(u.km / u.s) + v0  # km/s
    ev_t = np.empty(v_t.shape) * u.km / u.s
    ev_t[:] = ev0  # km/s
    # Create temporary FITS files for testing
    filenames = []
    pix_size = 1 * u.arcsec.to(u.deg)

    hdu = fits.PrimaryHDU(v_t.value)
    hdu.header["CTYPE1"] = "RA---TAN"
    hdu.header["CTYPE2"] = "DEC--TAN"
    hdu.header["CUNIT1"] = "deg"
    hdu.header["CUNIT2"] = "deg"
    hdu.header["CRPIX1"] = half_size
    hdu.header["CRPIX2"] = half_size
    # somewhere in Perseus
    hdu.header["CRVAL1"] = (15 * 3.5, "deg")
    hdu.header["CRVAL2"] = (31 + 1.0 / 3.0, "deg")
    hdu.header["CDELT1"] = (-1 * pix_size, "deg")
    hdu.header["CDELT2"] = (pix_size, "deg")
    hdu.header["BUNIT"] = ("km / s", "Vlsr")
    # like a single dish
    hdu.header["BMAJ"] = 2 * hdu.header["CDELT2"]
    hdu.header["BMIN"] = hdu.header["BMAJ"]
    hdu.header["BPA"] = 45.0
    filename = tmp_path / "test_vc.fits"
    hdu.writeto(filename)
    filenames.append(str(filename))

    filename = tmp_path / "test_evc.fits"
    # data2 = data * 0.0 + 0.01
    hdu2 = fits.PrimaryHDU(ev_t.value, header=hdu.header)
    hdu2.writeto(filename)
    filenames.append(str(filename))

    return filenames


def test_vfit(velocity_file) -> None:

    vc, hd = fits.getdata(velocity_file[0], header=True)
    evc = fits.getdata(velocity_file[1], header=False)
    x_t = np.arange(hd["NAXIS1"]) * hd["CDELT1"] * (u.deg)
    y_t = np.arange(hd["NAXIS2"]) * hd["CDELT2"] * (u.deg)
    xv, yv = np.meshgrid(x_t, y_t, indexing="xy")
    result_all = velocity_tools.vgrad.vfit(
        xv, yv, vc * u.km / u.s, evc * u.km / u.s, distance=my_distance
    )
    conv_grad = (1 * u.pc.to(u.au) / 10.0 * u.arcsec).to(u.deg) / u.pc
    input_grad = conv_grad * np.sqrt(gx**2 + gy**2)
    input_PA = np.rad2deg(np.arctan2(gx.value, gy.value))

    assert pytest.approx(result_all["grad"]) == input_grad.value
    assert pytest.approx(result_all["posang"]) == input_PA
    assert pytest.approx(result_all["vc"], abs=0.05) == v0.value


def test_vfit_local(velocity_file) -> None:

    vc, hd = fits.getdata(velocity_file[0], header=True)
    evc = fits.getdata(velocity_file[1], header=False)
    x_t = np.arange(hd["NAXIS1"]) * hd["CDELT1"] * (u.deg)
    y_t = np.arange(hd["NAXIS2"]) * hd["CDELT2"] * (u.deg)
    xv, yv = np.meshgrid(x_t, y_t, indexing="xy")
    result_all = velocity_tools.vgrad.vfit_local(
        xv, yv, vc * u.km / u.s, evc * u.km / u.s, distance=my_distance, nmin=5
    )
    conv_grad = (1 * u.pc.to(u.au) / 10.0 * u.arcsec).to(u.deg) / u.pc
    input_grad = conv_grad * np.sqrt(gx**2 + gy**2)
    input_PA = np.rad2deg(np.arctan2(gx.value, gy.value))

    assert pytest.approx(result_all["grad"]) == input_grad.value
    assert pytest.approx(result_all["posang"]) == input_PA
    assert pytest.approx(result_all["vc"], abs=0.05) == v0.value


def test_vfit_image(velocity_file) -> None:
    result_all = velocity_tools.vgrad.vfit_image(
        velocity_file[0], velocity_file[1], distance=my_distance, n_oversample=2
    )
    conv_grad = (1 * u.pc.to(u.au) / 10.0 * u.arcsec).to(u.deg) / u.pc
    input_grad = conv_grad * np.sqrt(gx**2 + gy**2)
    input_PA = np.rad2deg(np.arctan2(gx.value, gy.value))

    assert pytest.approx(result_all["grad"]) == input_grad.value
    assert pytest.approx(result_all["posang"]) == input_PA
    assert pytest.approx(result_all["vc"].value, abs=0.05) == v0.value

    assert pytest.approx(result_all["header"]["VG"]) == input_grad.value
    assert pytest.approx(result_all["header"]["VG_PA"]) == input_PA
    assert pytest.approx(result_all["header"]["VG_vc"], abs=0.05) == v0.value

    with pytest.raises(
        FileNotFoundError, match="File not found: nonexistent_file.fits"
    ):
        # Test with a non-existent file
        velocity_tools.vgrad.vfit_image(
            "nonexistent_file.fits",
            velocity_file[1],
            distance=my_distance,
            n_oversample=2,
        )
    with pytest.raises(
        FileNotFoundError, match="File not found: nonexistent_file.fits"
    ):
        # Test with a non-existent file
        velocity_tools.vgrad.vfit_image(
            velocity_file[0],
            "nonexistent_file.fits",
            distance=my_distance,
            n_oversample=2,
        )



import numpy as np
import astropy.units as u
from .coordinate_offsets import generate_offsets
from radio_beam import Beam
from astropy.convolution import convolve
from skimage.morphology import erosion

class result_container:
    """ Function to create class container
    """
    pass

# @u.quantity_input
def convolve_Vlsr(V_lsr:np.ndarray, header, dilation:bool=False) -> np.ndarray:
    """ 
    It Convolves a pure theoretical Vlsr map with a requested beam.
    The beam is setup using the FITS header of the expected observation.
    The convolution would mimick (at least at first order) the effect of a 
    finite beam size in the observations.
    The header is also used to convert the beam into pixel units before 
    convolving the map

    param :
    Vlsr : image with Velocity map to be convolved. It handles astropy.units.
    header : FITS header with beam and pixel size information
    dilation : bool, optional  
        If dilation is True, then mask all convolved data beyond
        one beam.
    """
    pixscale = np.abs(header['cdelt1'])*u.Unit(header['cunit1'])
    my_beam = Beam.from_fits_header(header)
    my_beam_kernel = my_beam.as_kernel(pixscale)
    if dilation == False:
        return convolve(V_lsr, my_beam_kernel, boundary='fill', fill_value=np.nan)
    else:
        v_convolved = convolve(V_lsr, my_beam_kernel, boundary='fill', fill_value=np.nan)
        nimage = np.ceil(np.abs(header['BMAJ'] / header['CDELT1'])).astype(int)
        if nimage % 2 == 0:
            nimage += 1
        # creates box with zeros with the size of the major axis, 
        # and place a 1 at the center
        center = (nimage - 1) // 2
        box_beam = np.zeros((nimage, nimage))
        box_beam[center, center] = 1
        # convolve box with beam and then threshold it to half-power
        beam_shape = convolve(box_beam, my_beam_kernel)
        beam_footprint = ((beam_shape / beam_shape.max()) > 0.5) + 0
        bad = erosion(np.isnan(V_lsr), beam_footprint)
        v_convolved[bad] = np.nan
        return v_convolved 

@u.quantity_input#(R_out=u.au, Mstar=u.Msun, Vc=u.km/u.s, inclination)
def keplerian_field(radius_2d:u.au, Pangle_2d:u.deg, 
                    Mstar:u.Msun =1*u.Msun, 
                    inclination:u.deg = 30*u.deg, 
                    R_out:u.au = 300*u.au) -> u.km/u.s:
    """
    Keplerian velocity field, for a star of mass=Mstar, inclination angle with
    respect of the sky of inclination.
    The position in the disk is described in polar coordinates, radius and angle.
    The outer disk radius is Rout.
    It makes full use of astropy.units.

    param :
    radius : radius in distance units (e.g. u.au or u.pc)
    angle : position angle with respect to major axis

    Mstar : central stellar mass (e.g. u.Msun)
    Vlsr : velocity of the star (u.km/u.s)
    inclination : with units (e.g u.deg or u.rad)
    R_out : Maximum radius of the disk. Position outside this radius are blanked
    """
    Kep_velo = 29.78 * np.sqrt(Mstar/u.Msun / (radius_2d/u.au)) \
               * np.abs(np.sin(inclination)) * np.cos(Pangle_2d)*u.km/u.s
    Kep_velo[radius_2d > R_out] = np.nan

    return Kep_velo

@u.quantity_input
def generate_Vlsr(header, ra0:u.deg, dec0:u.deg, frame:str='fk5',
    PA_Angle:u.deg=142.*u.deg, inclination:u.deg=42.*u.deg, distance:u.pc=110.0*u.pc,
    R_out:u.au=300.*u.au, Mstar:u.Msun=2.2*u.Msun, Vc:u.km/u.s=5.2*u.km/u.s) -> result_container:
    """
    Keplerian velocity field, for a star of mass=Mstar, inclination angle with 
    respect of the sky of inclination.
    The position in the disk is described in polar coordinates, radius and angle.
    The central velocity of the star is Vlsr, 
    and the maximum outer disk radius is Rout
    It makes full use of astropy.units.

    param :
    radius : radius in distance units (e.g. u.au or u.pc)
    angle : position angle with respect to major axis

    Mstar : central stellar mass (e.g. u.Msun)
    Vlsr : velocity of the star (u.km/u.s)
    inclination : with units (e.g u.deg or u.rad)
    R_out : Maximum radius of the disk. Position outside this radius are blanked
    """
    #
    results = generate_offsets(header, ra0, dec0, frame=frame,
                     pa_angle=PA_Angle, inclination=inclination)
    # center = SkyCoord(ra0, dec0, frame='fk5')
    # Load WCS 
    # w = wcs.WCS(header)
    # Create xy array and then coordinates
    # x = np.arange(header['naxis1'])
    # y = np.arange(header['naxis2'])
    #
    # epsilon will be determined as the pixel size
    epsilon = (np.abs(header['cdelt1'])*u.deg).to('',
                equivalencies=u.dimensionless_angles())*distance.to(u.au)
    # xx, yy = np.meshgrid(x, y)
    # world = w.wcs_pix2world(xx.flatten(), yy.flatten(), 0)

    # radec = SkyCoord(world[0]*u.deg, world[1]*u.deg, frame='fk5')
    # radec_off = radec.transform_to(center.skyoffset_frame())
    #
    # Ra = Lon, Dec = Lat
    #
    # lon = radec_off[:].lon
    # lat = radec_off[:].lat
    # lon.shape = xx.shape
    # lat.shape = xx.shape
    # This plots the offset in x and y
    # This is all good now
    # c, s = np.cos(PA_Angle), np.sin(PA_Angle)
    # Rotate the axes
    # Ra = Lon, Dec = Lat
    # lat_PA = c*lat + s*lon
    # lon_PA = -s*lat + c*lon
    # Deprojection 
    # Major axis in in RA direction
    # lon_PA /= np.cos(inclination)
    # dep_angle = np.sqrt(lat_PA**2 + lon_PA**2)
    # angle_PA = np.arctan2(lon_PA, lat_PA)
    #
    # dep_angle
    angle_PA = results.theta
    dep_radius = np.clip(results.r.to('',
                         equivalencies=u.dimensionless_angles())*distance.to(u.au),
                         epsilon.to(u.au), None)
    Kep_velo = keplerian_field(dep_radius, angle_PA, 
                    Mstar=Mstar, 
                    inclination=inclination,
                    R_out=R_out) 
    # Kep_velo = 29.78 * np.sqrt(Mstar/u.Msun / (dep_radius/u.au)) \
    #            * np.abs(np.sin(inclination)) * np.cos(angle_PA)*u.km/u.s
    Kep_velo += Vc
    # Kep_velo[dep_radius > R_out] = np.nan
    #
    # if do_plot:
    #     # Plot
    #     axes_extent = [(results.pre_lon.to(u.arcsec).value).max(),
    #                    (results.pre_lon.to(u.arcsec).value).min(),
    #                    (results.pre_lat.to(u.arcsec).value).min(),
    #                    (results.pre_lat.to(u.arcsec).value).max()]
    #     plt.ion()
    #     plt.figure(figsize=(12, 6))
    #     plt.subplot(1, 2, 1)
    #     plt.imshow(dep_radius.value, origin='lower', interpolation='none',
    #                extent=axes_extent)
    #     dx_line = (results.pre_lat.to(u.arcsec).value).max() * 0.2
    #     dv_disk = (np.nanmax(np.abs(Kep_velo - Vc))).value
    #     plt.contour(dep_radius.value,
    #                 np.linspace(np.nanmin(dep_radius.value),
    #                             np.nanmax(dep_radius.value), num=10),
    #                 colors='k', extent=axes_extent)
    #     plt.plot([0.0, dx_line*np.sin(PA_Angle)],
    #              [0.0, dx_line*np.cos(PA_Angle)], color='gray')
    #     plt.plot([0.0, dx_line*np.sin(PA_Angle+90.*u.deg)],
    #              [0.0, dx_line*np.cos(PA_Angle+90.*u.deg)], color='red')
    #     plt.title('Deprojected radius, $r_{PA,i}$')
    #     #
    #     plt.subplot(1, 2, 2)
    #     plt.imshow(Kep_velo.value, origin='lower', cmap='RdYlBu_r',
    #                interpolation='none', extent=axes_extent)
    #     plt.contour(Kep_velo.value,
    #                 np.linspace(Vc.value - dv_disk,
    #                             Vc.value + dv_disk,
    #                             #np.nanmin(Kep_velo.value),
    #                             #np.nanmax(Kep_velo.value),
    #                             num=10),
    #                 colors='k', extent=axes_extent)
    #     plt.title('Projected $V_{Kep}$')
    # Store results on class
    # results = result_container()
    results.r = dep_radius
    # results.theta = angle_PA
    results.v = Kep_velo
    # results.lat = lat_PA
    # results.lon = lon_PA
    return results



import astropy.units as u
import numpy as np
from astropy.io import fits
import velocity_tools.coordinate_offsets as coordinate_offsets

import pytest


def test_generate_offsets() -> None:
    #
    # check header is valid
    data = np.ones(shape=(11, 11))
    hdu = fits.PrimaryHDU(data)
    ra0 = 3.5 * u.deg
    dec0 = 30 * u.deg
    cdelt = 10 * u.arcsec.to(u.deg)
    hdu.header["CRVAL1"] = (ra0.value, "deg")
    hdu.header["CRVAL2"] = (dec0.value, "deg")
    hdu.header["CRPIX1"] = 5.0
    hdu.header["CRPIX2"] = 5.0
    hdu.header["CDELT1"] = -cdelt
    hdu.header["CDELT2"] = cdelt
    hdu.header["CTYPE1"] = "RA---TAN"
    hdu.header["CTYPE2"] = "DEC--TAN"

    results = coordinate_offsets.generate_offsets(
        hdu.header, ra0, dec0, frame="fk5", pa_angle=0 * u.deg, inclination=0 * u.deg
    )
    # Deprojection
    # Major axis in in Lon direction
    dxi = np.arange(-5, 6) * -cdelt * u.deg
    dyi = np.arange(-5, 6) * cdelt * u.deg

    dx, dy = np.meshgrid(dxi, dyi)

    # ad  more comparisons
    assert results.lat.shape == (11, 11)
    assert (dx == results.lat).all
    assert (dy == results.lon).all
    assert (np.hypot(dx, dy) == results.r).all


def test_average_profile() -> None:
    from velocity_tools.coordinate_offsets import average_profile

    # Generate some data
    def function_profile(x):
        return 10 + x**-2
    n_sample = 1000
    x_axis = 1 + 10 * np.random.rand(n_sample)
    y_axis = function_profile(x_axis) + np.random.rand(n_sample) * 0.1

    # Calculate the average profile
    xbin, ybin, dxbin, dybin = coordinate_offsets.average_profile(x_axis, y_axis, dx=1.0)
    assert len(xbin) == len(ybin)
    assert len(xbin) == len(dxbin)
    assert len(xbin) == len(dybin)
    assert pytest.approx(ybin, rel=0.05) == function_profile(xbin)


# %run stream_lines.py
import matplotlib.pyplot as plt
import numpy as np
import astropy.units as u
from astropy.constants import G
from velocity_tools import stream_lines

plt.ion()

theta0_1 = 30*u.deg
theta0_2 = 70*u.deg

Omega0 = 4*(1.7 * u.km/u.s/u.pc).decompose()
r0 = 1e4*u.au
Mstar = 0.5*u.Msun
Omega1 = (np.sqrt(G * Mstar * 300*u.au) / r0**2).decompose()

Rc = R_cent(M=Mstar, Omega=Omega1, r0=r0)
r = np.arange(r0.value, Rc.value, step=-10) * u.au
angles = np.linspace(0, 2 * np.pi, 100)


theta1 = stream_line(r, M=Mstar, theta0=theta0_1, 
    Omega=Omega1, r0=r0)
# theta1_p = stream_line(r, M=Mstar, theta0=theta0_1+10*u.deg,
#     Omega=Omega1, r0=r0)
# theta1_m = stream_line(r, M=Mstar, theta0=theta0_1-10*u.deg, 
#     Omega=Omega1, r0=r0)
theta2 = stream_line(r, M=Mstar, theta0=theta0_2, 
    Omega=Omega1, r0=r0)

dphi2 = stream_lines.dphi(theta2, theta0=theta0_2)
dphi1 = dphi(theta1, theta0=theta0_1)

# theta2_p = stream_line(r, M=Mstar, theta0=theta0_2+10*u.deg,
#     Omega=Omega1, r0=r0)
# theta2_m = stream_line(r, M=Mstar, theta0=theta0_2-10*u.deg, 
#     Omega=Omega1, r0=r0)

phi1 = 15 * u.deg
dphi_f = 15 * u.deg

z1 = r * np.cos(theta1)
y1 = r * np.sin(theta1) * np.sin(phi1 + dphi1)
x1 = r * np.cos(theta1) * np.cos(phi1 + dphi1)

y1_p = r * np.sin(theta1) * np.sin(phi1 + dphi1 + dphi_f)
x1_p = r * np.cos(theta1) * np.cos(phi1 + dphi1 + dphi_f)

y1_m = r * np.sin(theta1) * np.sin(phi1 + dphi1 - dphi_f)
x1_m = r * np.cos(theta1) * np.cos(phi1 + dphi1 - dphi_f)

z2 = r * np.cos(theta2)
y2 = r * np.sin(theta2) * np.sin(phi1 + dphi2)
x2 = r * np.cos(theta2) * np.cos(phi1 + dphi2)

y2_p = r * np.sin(theta2) * np.sin(phi1 + dphi2 + dphi_f)
x2_p = r * np.cos(theta2) * np.cos(phi1 + dphi2 + dphi_f)

y2_m = r * np.sin(theta2) * np.sin(phi1 + dphi2 - dphi_f)
x2_m = r * np.cos(theta2) * np.cos(phi1 + dphi2 - dphi_f)


# x1_m = r * np.cos(theta1_m)
# x1_p = r * np.cos(theta1_p)
# x2 = r * np.cos(theta2)
# x2_m = r * np.cos(theta2_m)
# x2_p = r * np.cos(theta2_p)
# xl_1 = r * np.cos(theta1[-1])
xl_2 = r * np.cos(theta2[-1])
#
# y1 = r * np.sin(theta1)
# y1_m = r * np.sin(theta1_m)
# y1_p = r * np.sin(theta1_p)
# y2 = r * np.sin(theta2)
# y2_m = r * np.sin(theta2_m)
# y2_p = r * np.sin(theta2_p)
# yl_1 = r * np.sin(theta1[-1])
# yl_2 = r * np.sin(theta2[-1])

plt.close()
plt.plot(x1, y1, color='r', marker='o', markersize=3)
plt.plot(x1_m, y1_m, color='r', marker='o', markersize=3)
plt.plot(x1_p, y1_p, color='r', marker='o', markersize=3)
# plt.plot(xl_1, yl_1, color='0.9')
plt.plot(x2, y2, color='b', marker='o', markersize=3)
plt.plot(x2_m, y2_m, color='b', marker='o', markersize=3)
plt.plot(x2_p, y2_p, color='b', marker='o', markersize=3)
# plt.plot(xl_2, yl_2, color='0.9')
plt.plot(Rc * np.cos(angles), Rc * np.sin(angles), color='r')
plt.axis('equal')
plt.xlim(-r.max().value, r.max().value)
plt.ylim(-r.max().value, r.max().value)
plt.tight_layout()



import numpy as np
import astropy.units as u
from scipy import optimize
from astropy.constants import G


#
# Implementation of stream lines using the prescription from
# Mendoza et al. (2009)  doi:10.1111/j.1365-2966.2008.14210.x
#

def v_k(radius, mass=0.5 * u.Msun):
    """
    Velocity term that is repeated in all velocity component.
    It corresponds to v_k in Mendoza+(2009)
    """
    return np.sqrt(G * mass / radius).to(u.km / u.s)


def r_cent(mass=0.5 * u.Msun, omega=1e-14 / u.s, r0=1e4 * u.au):
    """
    Centrifugal radius or disk radius in the Ulrich (1976)'s model.
    r_u in Mendoza's nomenclature.

    :param mass: Central mass for the protostar
    :param omega: Angular speed at the r0 radius
    :param r0: Initial radius of the streamline
    :return:
    """
    return (r0 ** 4 * omega ** 2 / (G * mass)).to(u.au)


def theta_abs(theta, r_to_rc=0.1, theta0=np.radians(30), ecc=1.,
              orb_ang=90*u.deg):
    """
    function to determine theta numerically by finding the root of a function
    This is equation (9) in Mendoza+(2009)

    :param theta: angle in radians
    :param r_to_rc: radius in units of the centrifugal radius
    :param theta0: Initial angle of the streamline
    :param ecc: eccentricity of the orbit (equation 6)
    :param orb_ang: angle in the orbital motion (equation 7)
    :return: returns the difference between the radius and the predicted one,
           a value of 0 corresponds to a proper streamline
    """
    cos_ratio = np.cos(theta) / np.cos(theta0)
    if cos_ratio > 1.:
        print('theta0={0}, theta_try={1} --> bad arccos calculation'.format(theta0, theta))
        return np.nan
    xi = np.arccos(cos_ratio) + orb_ang.to(u.rad).value
    geom = np.sin(theta0)**2 / (1 - ecc * np.cos(xi))
    return np.abs(r_to_rc - geom)


def get_dphi(theta, theta0=np.radians(30)):
    """
    Gets the difference in Phi.

    :param theta: radians
    :param theta0: radians

    :return difference in Phi angle, in radians
    """
    return np.arccos(np.tan(theta0) / np.tan(theta))


def stream_line(r, mass=0.5 * u.Msun, r0=1e4 * u.au, theta0=30 * u.deg,
                omega=1e-14 / u.s, v_r0=0 * u.km / u.s):
    """
    It calculates the stream line following Mendoza et al. (2009)
    It takes the radial velocity and rotation at the streamline
    initial radius and it describes the entire trajectory.

    :param r:
    :param mass:
    :param r0:
    :param theta0:
    :param phi0:
    :param omega:
    :param v_r0: Initial radial velocity
    :return: theta
    """
    # convert theta0 into radians
    rad_theta0 = theta0.to(u.rad).value
    rc = r_cent(mass=mass, omega=omega, r0=r0)
    print("rc={0}".format(rc))
    theta = np.zeros_like(r.value) + np.nan
    # mu and nu are dimensionless
    mu = (rc / r0).decompose().value
    nu = (v_r0 * np.sqrt(rc / (G * mass))).decompose().value
    # epsilon is the dimensionless energy
    epsilon = nu**2 + mu**2 * np.sin(theta0)**2 - 2 * mu
    ecc = np.sqrt(1 + epsilon * np.sin(theta0)**2)
    orb_ang = np.arccos((1 - mu * np.sin(theta0)**2) / ecc)
    # the first element in the streamline is the starting point
    theta[0] = rad_theta0
    # Initial guess at largest radius is theta0 +- initguess towards the midplane
    deltar = np.amin(np.abs(np.roll(r,1) - r))
    print(deltar)
    # we use a constant of 6e-5 for an epsilon of 0.01 km/s
    # this result will be in radians
    tol = (6.e-5 * deltar * omega / (v_r0+ 0.1 * u.km/u.s)).decompose().value
    # the initial guess will be 10 times the tolerance for now, in testing
    initguess = 10 * tol
    print('tolerance ', tol)
    if rad_theta0 < np.radians(90):
        theta_i = rad_theta0 + initguess
        theta_bracket = [(rad_theta0, np.pi/2.)]
    else:
        theta_i = rad_theta0 - initguess
        theta_bracket = [(np.pi/2., rad_theta0)]
    for ind in np.arange(1, len(r)):
        r_i = (r[ind] / rc).decompose().value
        if r_i > 0.5:
            # print('initial guess of theta_i = {0}'.format(theta_i))
            # result = optimize.minimize(theta_abs, theta_i,
            #                            bounds=theta_bracket,
            #                            args=(r_i, rad_theta0, ecc, orb_ang))
            # By default, when minimize receives bounds and no constrains,
            # it uses the L-BFGS-B method:
            # ftol is the tolerance in the function evaluation
            # "The iteration stops when (f^k - f^{k+1})/max{|f^k|,|f^{k+1}|,1} <= ftol"
            # gtol corresponds to the parameter pgtol in fmin_l_bfgs_b
            # "The iteration will stop when max{|proj g_i | i = 1, ..., n} <= gtol"
            # eps corresponds to the absolute step size used for numerical approximation of the jacobian via forward differences.
            options_dict = {'gtol': tol/10., 'eps': tol, 'ftol': tol}
            result = optimize.minimize(theta_abs, theta_i,
                                       bounds=theta_bracket,
                                       args=(r_i, rad_theta0, ecc, orb_ang),
                                       options=options_dict)
            theta_i = result.x
            # These prints are to diagnose if the minimization is converging
            # print(ind, result.success)
            # print(result.message, result.status, result.nit)
            theta[ind] = theta_i
    return theta * u.rad


def stream_line_vel(r, theta, mass=0.5*u.Msun, r0=1e4*u.au, theta0=30*u.deg,
                omega=1e-14/u.s, v_r0=0*u.km/u.s):
    """
    It calculates the velocity along the stream line following Mendoza+(2009)
    It takes the radial velocity and rotation at the streamline
    initial radius and it describes the entire trajectory.

    :param theta:
    :param r:
    :param mass:
    :param r0:
    :param theta0:
    :param phi0:
    :param omega:
    :param v_r0: Initial radial velocity
    :return: v_r, v_theta, v_phi in units of km/s
    """
    rc = r_cent(mass=mass, omega=omega, r0=r0)
    r_to_rc = (r / rc).decompose().value
    v_k0 = v_k(rc, mass=mass)
    # mu and nu are dimensionless
    mu = (rc / r0).decompose().value
    nu = (v_r0 * np.sqrt(rc / (G * mass))).decompose().value
    epsilon = nu**2 + mu**2 * np.sin(theta0)**2 - 2 * mu
    ecc = np.sqrt(1 + epsilon*np.sin(theta0)**2)
    orb_ang = np.arccos((1 - mu * np.sin(theta0)**2) / ecc)
    cos_ratio = np.cos(theta) / np.cos(theta0)
    xi = np.arccos(cos_ratio) + orb_ang.to(u.rad)#.value
    #
    v_r_all = -ecc * np.sin(theta0) * np.sin(xi) / r_to_rc /(1 - ecc*np.cos(xi))
    v_theta_all = np.sin(theta0) / np.sin(theta) / r_to_rc \
                  * np.sqrt(np.cos(theta0)**2 - np.cos(theta)**2)
    v_phi_all = np.sin(theta0)**2 / np.sin(theta) / r_to_rc

    return v_r_all * v_k0, v_theta_all * v_k0, v_phi_all * v_k0


def rotate_xyz(x, y, z, inc=30 * u.deg, pa=30 * u.deg):
    """
    Rotate on inclination and PA
    x-axis and y-axis are on the plane on the sky,
    z-axis is the

    Rotation around x is inclination angle
    Rotation around y is PA angle

    Using example matrices as described in:
    https://en.wikipedia.org/wiki/3D_projection

    :param x: cartesian x-coordinate, in the direction of decreasing RA
    :param y: cartesian y-coordinate, in the direction away of the observer
    :param z: cartesian z-coordinate, in the direction of increasing Dec.
    :param inc: Inclination angle. 0=no change
    :param pa: Change the PA angle. Measured from North due East.
    :return: new x, y, and z-coordinates as observed on the sky, with the
    same units as the input ones.

    """
    xyz = np.column_stack((x, y, z))
    rot_inc = np.array([[1, 0, 0],
                        [0, np.cos(inc), np.sin(inc)],
                        [0, -np.sin(inc), np.cos(inc)]])
    rot_pa = np.array([[np.cos(pa), 0, -np.sin(pa)],
                       [0, 1, 0],
                       [np.sin(pa), 0, np.cos(pa)]])
    xyz_new = rot_pa.dot(rot_inc.dot(xyz.T))
    return xyz_new[0], xyz_new[1], xyz_new[2]


@u.quantity_input
def xyz_stream(mass=0.5*u.Msun, r0=1e4*u.au, theta0=30*u.deg,
               phi0=15*u.deg, omega=1e-14/u.s, v_r0=0*u.km/u.s,
               inc=0*u.deg, pa=0*u.deg, rmin=None, deltar=1*u.au):
    """
    it gets xyz coordinates and velocities for a stream line.
    They are also rotated in PA and inclination along the line of sight.
    This is a wrapper around stream_line() and rotate_xyz()

    Spherical into cartesian transformation is done for position and velocity
    using:
    https://en.wikipedia.org/wiki/Vector_fields_in_cylindrical_and_spherical_coordinates

    :param mass: Central mass
    :param r0: Initial radius of streamline
    :param theta0: Initial polar angle of streamline
    :param phi0: Initial azimuthal angle of streamline
    :param omega: Angular rotation. (defined positive)
    :param v_r0: Initial radial velocity of the streamline
    :param inc: inclination with respect of line-of-sight, inc=0 is an edge-on-disk
    :param pa: Position angle of the rotation axis, measured due East from North. This is usually estimated from the outflow PA, or the disk PA-90deg.
    :param rmin: smallest radius for calculation
    :param deltar: spacing between two consecutive radii in the sampling of the streamer, in au
    :return:
    """
    rc = r_cent(mass=mass, omega=omega, r0=r0)
    if rc > r0:
        print('Centrifugal radius is larger than start of streamline')
    r = np.arange(r0.to(u.au).value, rc.to(u.au).value*0.5, step=-1*deltar.value) * u.au
    theta = stream_line(r, mass=mass, r0=r0, theta0=theta0,
                        omega=omega, v_r0=v_r0)
    d_phi = get_dphi(theta, theta0=theta0)
    phi = phi0 + d_phi
    #
    v_r, v_theta, v_phi = stream_line_vel(r, theta, mass=mass, r0=r0,
                                          theta0=theta0, omega=omega, v_r0=v_r0)
    v_x = v_r * np.sin(theta) * np.cos(phi) \
          + v_theta * np.cos(theta) * np.cos(phi) \
          - v_phi * np.sin(phi)
    v_y = v_r * np.sin(theta) * np.sin(phi) \
          + v_theta * np.cos(theta) * np.sin(phi) \
          + v_phi * np.cos(phi)
    v_z = v_r * np.cos(theta) \
          - v_theta * np.sin(theta)
    # Convert from spherical into cartesian coordinates
    x = r * np.sin(theta) * np.cos(phi)
    y = r * np.sin(theta) * np.sin(phi)
    z = r * np.cos(theta)
    if rmin is not None:
        gd_rmin = (r > rmin)
        if gd_rmin.sum() > 0:
            return rotate_xyz(x[gd_rmin], y[gd_rmin], z[gd_rmin], inc=inc, pa=pa),\
                rotate_xyz(v_x[gd_rmin], v_y[gd_rmin], v_z[gd_rmin], inc=inc, pa=pa)
        else:
            return [np.nan], [np.nan], [np.nan], [np.nan], [np.nan], [np.nan]
    else:
        return rotate_xyz(x, y, z, inc=inc, pa=pa), \
               rotate_xyz(v_x, v_y, v_z, inc=inc, pa=pa)



# %run stream_lines.py
import numpy as np
import astropy.units as u
import matplotlib.pyplot as plt

# Initial test: change in PA for vector = showin x- and y-axes
x_x = 1.
y_x = 0.
z_x = 0.
x_y = 0.
y_y = 1.
z_y = 0.

x_x_new, y_x_new, z_x_new = rotate_xyz(x_x, y_x, z_x, inc=0*u.deg, PA=45*u.deg)
x_y_new, y_y_new, z_y_new = rotate_xyz(x_y, y_y, z_y, inc=0*u.deg, PA=45*u.deg)

# Plot x- and y-axes only
plt.ion()
plt.close()
fig1, ax1 = plt.subplots(figsize=(7,7))
ax1.plot([0, x_x], [0, y_x], color='red')
ax1.plot([0, x_y], [0, y_y], color='red')
ax1.plot([0, x_x_new[0]], [0, y_x_new[0]], color='blue')
ax1.plot([0, x_y_new[0]], [0, y_y_new[0]], color='blue')
ax1.axis('equal')
ax1.set_xlabel('x')
ax1.set_ylabel('y')


# test: change in inclination for vector = showin x- and y-axes
x_x = 1.
y_x = 0.
z_x = 0.
x_y = 0.
y_y = 1.
z_y = 0.

ang_c = np.linspace(0, 2*np.pi, 100)
x_circ = np.cos(ang_c)
y_circ = np.sin(ang_c)
z_circ = np.cos(ang_c) * 0.0
x_circ_new, y_circ_new, z_circ_new = rotate_xyz(x_circ, y_circ, z_circ, inc=60*u.deg, PA=30*u.deg)
# x_y_new, y_y_new, z_y_new = rotate_xyz(x_y, y_y, z_y, inc=60*u.deg, PA=0*u.deg)

# Plot x- and y-axes only
fig2, ax2 = plt.subplots(figsize=(14,7), ncols=2)
ax2[0].plot([0, x_x], [0, y_x], color='red')
ax2[0].plot([0, x_y], [0, y_y], color='red')
ax2[0].plot( x_circ, y_circ, color='red')
ax2[0].plot( x_circ_new, y_circ_new, color='blue', ls=':')
ax2[0].axis('equal')
ax2[0].set_xlabel('x')
ax2[0].set_ylabel('y')

ax2[1].plot([0, z_x], [0, y_x], color='red')
ax2[1].plot([0, x_y], [0, y_y], color='red')
ax2[1].plot( z_circ, y_circ, color='red')
ax2[1].plot( z_circ_new, y_circ_new, color='blue', ls=':')
ax2[1].axis('equal')
ax2[1].set_xlabel('z')
ax2[1].set_ylabel('y')



import numpy as np
import astropy.units as u
from astropy import wcs
from astropy.coordinates import SkyCoord
import warnings


class result_container:
    """ Function to create class container
    """
    pass


def convolve_Vlsr(V_lsr, header):
    """ 
    It Convolves a pure theoretical Vlsr map with a requested beam.
    The beam is setup using the FITS header of the expected observation.
    The convolution would mimick (at least at first order) the effect of a 
    finite beam size in the observations.
    The header is also used to convert the beam into pixel units before 
    convolving the map

    param :
    Vlsr : image with Velocity map to be convolved. It handles astropy.units.
    header : FITS header with beam and pixel size information
    """
    warnings.warn("convolve_Vlsr has been moved to keplerian_field. Please use keplerian_field.convolve_Vlsr instead. It will be removed in the next release.", DeprecationWarning)
    from . import keplerian_field
    return keplerian_field.convolve_Vlsr(V_lsr, header, dilation=False)


def generate_Vlsr(radius, angle, inclination=42.*u.deg,
                  R_out=300.*u.au, Mstar=2.2*u.Msun, Vc=5.2*u.km/u.s) -> None:
    warnings.warn("generate_Vlsr has been deprecated. Please use keplerian_field.generate_Vlsr instead. It will be removed in the next release.", DeprecationWarning)
    return None

def generate_offsets(header, ra0, dec0, frame='fk5',
                     pa_angle=142.*u.deg, inclination=42.*u.deg):
    """
    Main geometry: major axis in deprojected lon variable, while lat is the minor axis

    :param header: FITS header of the file to be used
    :param ra0: RA of reference point for offset calculation (with units)
    :param dec0: Dec of reference point for offset calculation (with units)
    :param frame: coordinate frame for reference point (default='fk5')
    :param pa_angle: PA angle in deg
    :param inclination: inclination angle in deg
    :return: a structure with the radius and position angle in the deprojected
       coordinates, and the x and y offsets also in deprojected coordinates
    """
    #
    center = SkyCoord(ra0, dec0, frame=frame)
    # Load WCS
    w = wcs.WCS(header)
    # Create xy array and then coordinates
    xx, yy = np.meshgrid(np.arange(header['naxis1']),
                         np.arange(header['naxis2']))
    world = w.wcs_pix2world(xx.flatten(), yy.flatten(), 0)
    radec = SkyCoord(world[0]*u.deg, world[1]*u.deg, frame=frame)
    radec_off = radec.transform_to(center.skyoffset_frame())
    #
    # Ra = Lon, Dec = Lat
    #
    lon = radec_off[:].lon
    lat = radec_off[:].lat
    lon.shape = xx.shape
    lat.shape = xx.shape
    # Rotate the axes
    c, s = np.cos(pa_angle), np.sin(pa_angle)
    lat_pa = c*lat + s*lon
    lon_pa = -s*lat + c*lon
    # Deprojection
    # Major axis in in Lon direction
    lon_pa /= np.cos(inclination)
    # deprojected radius
    dep_angle = np.sqrt(lat_pa**2 + lon_pa**2)
    # deprojected angle
    angle_pa = np.arctan2(lon_pa, lat_pa)
    # Store results on class
    results = result_container()
    results.r = dep_angle
    results.theta = angle_pa
    results.pre_lat = lat
    results.pre_lon = lon
    results.lat = lat_pa
    results.lon = lon_pa
    return results


def mask_velocity(cube, v_map, v_width=1.0*u.km/u.s):
    """
    Returns a mask with the pixels in the channel mask within v_width of the
     expected velocity map (v_map). Values of 1.0s and 0.0s

    :param v_width:
    :param cube: SpectralCube cube object to work on.
    :param v_map: Centroid velocity map in velocity units.
    :return: Mask with shape the same as of the input cube,
      1=in the mask, 0=outside of the mask.
    """
    cube2 = cube.with_spectral_unit(u.km/u.s, velocity_convention='radio')
    vaxis = cube2.spectral_axis
    # Load Keplerian velocity model and give proper units
    vmask = np.zeros(cube2.shape)
    for ii in np.arange(0, vaxis.size):
        mask_i = np.abs(v_map - vaxis[ii]) < v_width
        vmask[ii, :, :] = mask_i
    return vmask.astype(np.float)


def average_profile(x, y, dx, dy=None, log=False, oversample=1.):
    """ 
    Averaging function to create a radial profile.

    It returns the average of y in the range [x, x+dx], where the bin size are 
    constant in linear space. 
    If log=True, then the bin sizes are constant in log space, and the average is 
    calculated in the range [log(x), log(x)+dx].

    oversample = 1. 
            This is the number to correct for the correlated number of pixesl within a beam. 
            Default is 1.
            Normal usage should be (beamsize/pixelsize)**2

    returns: xbin, ybin, dxbin, dybin
    xbin : the middle of the bin
    ybin : the mean of the y values in the given bin
    dxbin : the width of the bin
    dybin : the uncertainty on the mean value calculation. This is the standard deviation of the points.

    """
    if log == False:
        # Linear space
        xx = x
    else:
        # log space
        xx = np.log(x)

    xmin = np.min(xx)
    xmax = np.max(xx)
    n_bin = int(np.ceil((xmax-xmin)/dx))
    xbin = np.zeros(n_bin)
    dxbin = np.zeros(n_bin)
    ybin = np.zeros(n_bin)
    dybin = np.zeros(n_bin)
    for i in range(n_bin):
        idx = np.where((xx > xmin+dx*i) & (xx < xmin+dx*(i+1)))
        xbin[i] = xmin + dx*(i+0.5)
        #
        if dy is None:
            ybin[i] = np.average(y[idx])
            dybin[i] = np.std(y[idx]) / np.sqrt(y[idx].size / oversample)
        else:
            ybin[i] = np.average(y[idx], weights=1./dy[idx]**2)
            dybin[i] = 1. / np.sqrt(np.sum(1. / dy[idx]**2))
        dxbin[i] = dx*0.5
    if log == False:
        return xbin, ybin, dxbin, dybin
    else:
        return np.exp(xbin), ybin, dxbin, dybin



import keplerian_field
import matplotlib.pyplot as plt
import astropy.units as u
from astropy.io import fits
import numpy as np

# ra1 = ra0=15*(11+(33 + 25.318652/60.)/60.)*u.deg
# dec1=-1*(70+(11 + 41.23173/60.)/60.)*u.deg


def example_Vlsr(file_in='fits_files/test_file.fits'):
    """ How to use this thing
    """
    # Central coordinate
    ra0 = 15*(11+(33 + 25.321/60.)/60.)*u.deg
    dec0 = -1*(70+(11 + 41.209/60.)/60.)*u.deg
    header = fits.getheader(file_in)
    results = keplerian_field.generate_Vlsr(header, ra0, dec0,
                                            file_out='fits_files/test_Vc.fits', PA_Angle=(142.+180)*u.deg,
                                            # Inclination angle in Degrees
                                            inclination=42.*u.deg,
                                            # Distance in pc
                                            distance=110.02*u.pc,
                                            # Outer radius in au
                                            R_out=3.2*110.02*u.au,  # R_out=300.*u.au,
                                            # Stellar Mass in Solar Masses
                                            Mstar=2.2*u.Msun,
                                            # V_lsr
                                            Vc=5.64*u.km/u.s, do_plot=False)
    # Vc= 5.2*u.km/u.s, do_plot=False)
    conv_Vlsr = keplerian_field.convolve_Vlsr(results.v, header)
    plt.figure(figsize=(12, 6))
    plt.subplot(1, 3, 1)
    plt.imshow(results.r.value, origin='lowest', interpolation='none')
    plt.title('Deprojected radius, $r_{PA,i}$')
    plt.subplot(1, 3, 2)
    plt.imshow(results.theta.value, origin='lowest', interpolation='none')
    plt.title('Deprojected Angle, $theta$')
    plt.subplot(1, 3, 3)
    # plt.imshow(results.v.value, origin='lowest', cmap='RdYlBu_r', interpolation='none')
    plt.imshow(conv_Vlsr, origin='lowest',
               cmap='RdYlBu_r', interpolation='none')
    plt.title('Projected $V_{Kep}$')

    header2 = header
    header2['BUNIT'] = 'km/s'
    fits.writeto('fits_files/HD100546_Vc.fits',
                 results.v.value, header2, overwrite=True)
    header2 = header
    header2['BUNIT'] = 'km/s'
    fits.writeto('fits_files/HD100546_Vc_conv.fits',
                 conv_Vlsr, header2, overwrite=True)

    header2 = header
    header2['BUNIT'] = 'au'
    fits.writeto('fits_files/HD100546_Radius.fits',
                 results.r.value, header2, overwrite=True)

    header2 = header
    header2['BUNIT'] = 'deg'
    fits.writeto('fits_files/HD100546_theta.fits',
                 (results.theta.to(u.deg)).value, header2, overwrite=True)

    # header2=header
    # header2['BUNIT']='rad'
    # fits.writeto( 'fits_files/HD100546_theta.fits', results.theta.value, header2, overwrite=True)
    header2 = header
    header2['BUNIT'] = 'au'
    fits.writeto('fits_files/HD100546_lat.fits',
                 results.lat.value, header2, overwrite=True)

    header2 = header
    header2['BUNIT'] = 'au'
    fits.writeto('fits_files/HD100546_lon.fits',
                 results.lon.value, header2, overwrite=True)


if False:
    # Load data and convert cube into spectral axis of km/s
    example_Vlsr(file_in='fits_files/HD100546_masked_Vc.fits')
    from spectral_cube import SpectralCube
    cube = SpectralCube.read('fits_files/HD100546_12CO_mscale_cube_3D.fits')
    cube2 = cube.with_spectral_unit(u.km/u.s, velocity_convention='radio')
    vaxis = cube2.spectral_axis
    # Load keplerian velocity model and give proper units
    vkep = fits.getdata('fits_files/HD100546_Vc_conv.fits')*u.km/u.s
    vmask = np.zeros(cube2.shape)
    v_width = (1.0*u.km/u.s)
    for ii in np.arange(0, vaxis.size):
        mask_i = np.abs(vkep-vaxis[ii]) < v_width
        vmask[ii, :, :] = mask_i
    header_v = fits.getheader('fits_files/HD100546_12CO_mscale_cube_3D.fits')
    file_mask_out = 'fits_files/test_mask_1kms.fits'
    fits.writeto(file_mask_out, vmask.astype(
        np.float), header_v, overwrite=True)

    vmask2 = np.zeros(cube2.shape)
    v_width2 = (2.0*u.km/u.s)
    for ii in np.arange(0, vaxis.size):
        mask_i2 = np.abs(vkep-vaxis[ii]) < v_width2
        vmask2[ii, :, :] = mask_i2
    header_v = fits.getheader('fits_files/HD100546_12CO_mscale_cube_3D.fits')
    file_mask_out = 'fits_files/test_mask_2kms.fits'
    fits.writeto(file_mask_out, vmask2.astype(
        np.float), header_v, overwrite=True)

    file_mask_out = 'fits_files/test_mask_1kms.fits'
    my_mask = fits.getdata(file_mask_out)
    cube2_mask = cube2.with_mask(my_mask.astype(np.bool))
    m0 = cube2_mask.moment(order=0)
    m1 = cube2_mask.moment(order=1)
    m0.write('fits_files/test_mom0_masked_1kms.fits')
    m1.write('fits_files/test_mom1_masked_1kms.fits')

    file_mask_out = 'fits_files/test_mask_2kms.fits'
    my_mask = fits.getdata(file_mask_out)
    cube2_mask = cube2.with_mask(my_mask.astype(np.bool))
    m0 = cube2_mask.moment(order=0)
    m1 = cube2_mask.moment(order=1)
    m0.write('fits_files/test_mom0_masked_2kms.fits')
    m1.write('fits_files/test_mom1_masked_2kms.fits')
    # try:
    #     vmask[mask_i]=1
    # except:



import numpy as np
import sys
import os

import astropy.units as u
from astropy.io import fits


def vfit(x, y, v, ev, distance=300.0*u.pc):
    """
    .. py:function:: vfit(x, y, v, ev, [distance=300.0*u.pc])

    It calculates the velocity gradient to a group of velocity 
    measuments. 

    :param float x: The array with the offsets in x-axis (usually delta RA) 
    in angle units.
    :param float y: The array with the offsets in y-axis (usually delta Dec) 
    in angle units.
    :param float v: The array with the velocity measurements (usually Vlsr) 
    in velocity units.
    :param float ev: The array with the uncertainty in the velocity 
    measurements (usually Vlsr) in velocity units.
    :float distance: The distance to the region studied with distance 
    units. Default value is 300 pc.
    :return: structure with results from the fit. The gradient is in units 
    of km/s/pc, while the position angle is in degrees measured 
    East-from-North. Velocities are reported in km/s.
    """
    # TODO:
    # -Check input arrays have same length
    #
    if (np.count_nonzero(np.isnan(x))):  sys.exit('There are NaNs in x-axis')
    if (np.count_nonzero(np.isnan(y))):  sys.exit('There are NaNs in y-axis')
    if (np.count_nonzero(np.isnan(v))):  sys.exit('There are NaNs in v-axis')
    if (np.count_nonzero(np.isnan(ev))):  sys.exit('There are NaNs in velocity error')
    if (np.count_nonzero(ev == 0.0)):  sys.exit('There are 0.0 in velocity error')
    #
    if (distance <= 0):  sys.exit('distance cannot be negative')

    x_d = (x.to('', equivalencies=u.dimensionless_angles()) * distance).to(u.pc).value
    y_d = (y.to('', equivalencies=u.dimensionless_angles()) * distance).to(u.pc).value

    # Obtain total weight, and average (x,y,v) to create new variables (dx,dy,dv)
    # which provide a lower uncertainty in the fit.
    #
    v_d = v.to(u.km / u.s).value
    ev_d = ev.to(u.km / u.s).value
    wt = 1.0 / ev_d**2
    #
    suma = np.sum(wt)
    x_mean = np.sum(x_d * wt) / suma
    y_mean = np.sum(y_d * wt) / suma
    v_mean = np.sum(v_d * wt) / suma
    # relative units
    dx = (x_d - x_mean)
    dy = (y_d - y_mean)
    dv = (v_d - v_mean)
    #
    M = np.array([[np.sum(wt), np.sum(dx * wt), np.sum(dy * wt)], 
        [np.sum(dx * wt), np.sum(dx**2 * wt), np.sum(dx * dy * wt)], 
        [np.sum(dy * wt), np.sum(dx * dy * wt), np.sum(dy**2 * wt)]])
    try:
        covar = np.linalg.inv(M.T)
    except IOError:
        sys.exit('Singular matrix: no solution returned')
    coeffs = np.dot(covar, [[np.sum(dv * wt)], [np.sum(dx * dv * wt)],
                    [np.sum(dy * dv * wt)]])
    #
    errx = np.sqrt(covar[1, 1])
    erry = np.sqrt(covar[2, 2])
    #
    gx = coeffs[1][0]
    gy = coeffs[2][0]
    # best fit mean V_lsr and predicted velocities
    vc = coeffs[0][0] + v_mean
    vp = coeffs[0][0] + gx * dx + gy * dy
    grad = np.sqrt(gx**2 + gy**2)
    posang = np.rad2deg(np.arctan2(gx, gy))# + np.pi/2.)
    # chi^2_red
    red_chisq = np.sum((dv - vp)**2 * wt) / (len(dv) - 3.)

    vc_err = 0.
    # grad_err = np.sqrt((gx * errx)**2 + (gy * erry)**2) / grad
    # paerr = np.rad2deg(np.sqrt((gx / (gx**2 + gy**2))**2 * erry**2 +
    #                  (gy / (gx**2 + gy**2))**2 * errx**2))
    # Associated error including covariance
    grad_err = np.sqrt((gx * errx)**2 + (gy * erry)**2 
                     + 2 * gx * gy * covar[2, 1]) / grad
    
    paerr = np.rad2deg(np.sqrt((gx / (gx**2 + gy**2))**2 * erry**2 +    
                     (gy / (gx**2 + gy**2))**2 * errx**2 
                     - 2 * gx * gy / (gx**2 + gy**2)**2 * covar[2, 1]))
    chisq = red_chisq
    vp += v_mean

    results = {'grad': grad, 'grad_err': grad_err, 
               'posang': posang, 'paerr': paerr, 
               'chisq': chisq, 'vc': vc, 'vc_err': vc_err,
               'v_pred': vp}

    return results


def vfit_local(x, y, v, ev, distance=300.0*u.pc, width=45*u.arcsec, nmin=7.0):
    """
    .. py:function:: vfit_local(x, y, v, ev, [distance=300.0*u.pc],
    [width=45*u.arcsec], [nmin=7])

    It calculates the velocity gradient to a group of velocity 
    measuments. It uses a box of a requested width to determine which 
    pixels to use in the local calculation. The calculation is performed
    at every pixel, and it uses the vfit() function.



    nmin is the number of pixels, a value of 7 is a good value in the case
        of a Nyquist sampled dataset.
    :param float x: The array with the offsets in x-axis (usually delta RA) 
    in angle units.
    :param float y: The array with the offsets in y-axis (usually delta Dec) 
    in angle units.
    :param float v: The array with the velocity measurements (usually Vlsr) 
    in velocity units.
    :param float ev: The array with the uncertainty in the velocity 
    measurements (usually Vlsr) in velocity units.
    :float distance: The distance to the region studied with distance 
    units. Default value is 300 pc.
    :float width: The angular size of the box used in the pixel selection for 
    calculation, in units of angle. Default value is 45 arcsec.
    :float nmin: Minimum number of pixels in the requested box to perform the
    gradient fit. Default value is 7.
    :return: structure with results from all the fits.
    """
    
    grad = np.empty(x.shape)
    grad_err = np.empty(x.shape)
    posang = np.empty(x.shape)
    paerr = np.empty(x.shape)
    vc = np.empty(x.shape)
    chisq = np.empty(x.shape)
    #
    grad[:] = np.nan #
    grad_err[:] = np.nan #
    posang[:] = np.nan #
    paerr[:] = np.nan #
    vc[:] = np.nan #
    chisq[:] = np.nan

    for index, (x_i, y_i) in enumerate(zip(x, y)):
        # determine distances between all pixels to the one processing
        # and select all pixels within a radius of width.
        # dxy = np.sqrt((x - x_i)**2 + (y - y_i)**2)
        # gd = (dxy <= width)
        dx = np.abs(x - x_i)
        dy = np.abs(y - y_i)
        gd = (dx <= width * 0.5) & (dy <= width * 0.5)
        # determine the gradient only of there are enough pixels
        #
        if np.sum(gd) > nmin:
            res_grad = vfit(x[gd], y[gd], v[gd], ev[gd], distance=distance)
            grad[index] = res_grad['grad']
            grad_err[index] = res_grad['grad_err']
            posang[index] = res_grad['posang']
            paerr[index] = res_grad['paerr']
            vc[index] = res_grad['vc']
            chisq[index] = res_grad['chisq']

    # after all the calculations, we store the results and return them
    results = {'grad': grad, 'grad_err': grad_err, 
           'posang': posang, 'paerr': paerr, 
           'chisq': chisq, 'vc': vc, }
    return results


def vfit_image(file_vc, file_evc, distance=300.0*u.pc, width=None, 
        n_oversample=7):
    """
    .. py:function:: vfit_image(file_vc, file_evc, [distance=300.0*u.pc],
    [width=45*u.arcsec], [n_oversample=7])

    It loads the velocity and error in the velocity maps and calculates 
    the global velocity gradient and all the local velocity gradients 
    possible.
    It uses a box of a requested width to determine which 
    pixels to use in the local calculation.


    nmin is the number of pixels, a value of 7 is a good value in the case
        of a Nyquist sampled dataset.
    :param str file_vc: The filename to the FITS file with the centroid 
    velocity measurements. If no unit is present in the header, 
    it assumes km/s.
    :param str file_evc: The filename to the FITS file with the uncertainty 
    to the centroid velocity measurements. If no unit is present in the header, 
    it assumes km/s.
    :float distance: The distance to the region studied with distance 
    units. Default value is 300 pc.
    :float width: The angular size of the box used in the pixel selection for 
    calculation, in units of angle. Default value is 45 arcsec.
    :float n_oversample: Minimum number of pixels in the requested box to 
    perform the gradient fit (in a Nyquist sampled map). In the case of 
    oversampled maps (e.g., interferometers), the code takes the oversampling
    into account to correct for this. Default value is 7. 
    :return: structure with results from all the fits.
    """
    if (os.path.isfile(file_vc) == False):
        raise FileNotFoundError(f'File not found: {file_vc}')
    if (os.path.isfile(file_evc) == False):
        raise FileNotFoundError(f'File not found: {file_evc}')
    
    vc, hd = fits.getdata(file_vc, header=True)
    evc = fits.getdata(file_evc)
    if (hd['BUNIT'] == 'm/s') | (hd['BUNIT'] == 'm s-1'):
        unit_velo = 1e-3 * (u.km / u.s)
    else:
        unit_velo = u.km / u.s
    x = np.arange(hd['NAXIS1']) * hd['CDELT1'] * u.deg
    y = np.arange(hd['NAXIS2']) * hd['CDELT2'] * u.deg
    vc *= unit_velo
    evc *= unit_velo
    xv, yv = np.meshgrid(x, y, indexing='xy')
    gd = (np.isfinite(vc * evc) & (evc.value != 0))
    
    if width == None:
        width = 2 * hd['BMAJ'] * u.deg

    beam_pix = np.abs((hd['BMAJ'] * hd['BMIN']) / 
                     (4 * hd['CDELT1'] * hd['CDELT2']))
    result_all = vfit(xv[gd], yv[gd], vc[gd], evc[gd], distance=distance)
    result = vfit_local(xv[gd], yv[gd], vc[gd], evc[gd], distance=distance, 
        width=width, nmin=n_oversample * beam_pix)
    grad = np.empty(vc.shape) + np.nan
    grad_err = np.empty(vc.shape) + np.nan
    grad_PA = np.empty(vc.shape) + np.nan
    grad_PA_err = np.empty(vc.shape) + np.nan
    grad_vc = np.empty(vc.shape) + np.nan
    # v_pred = grad

    grad[gd] = result['grad']
    grad_err[gd] = result['grad_err']
    grad_PA[gd] = result['posang']
    grad_PA_err[gd] = result['paerr']
    grad_vc[gd] = result['vc']
    #
    # Store global fit results in header
    # 
    print(result_all)
    hd['VG'] = result_all['grad']
    hd['VGerr'] = result_all['grad_err']
    hd['VG_PA'] = result_all['posang']
    hd['VG_PAerr'] = result_all['paerr']
    hd['VG_vc'] = result_all['vc']
    results_image = {'grad': grad, 'grad_err': grad_err, 
           'posang': grad_PA, 'paerr': grad_PA_err, 
           'vc': vc, 'header':hd}
    return results_image



import pytest
import astropy.units as u
import numpy as np
from astropy.io import fits
from velocity_tools.keplerian_field import convolve_Vlsr
import velocity_tools.keplerian_field


@pytest.fixture
def generate_dummy_file(tmp_path) -> str:
    """ """
    file_in = tmp_path / "test_keplerian.fits"
    ra0 = 15 * (3 + (30 + 25.6 / 60.0) / 60.0)
    dec0 = 32 + (10 + 25.6 / 60.0) / 60.0
    # create dummy FITS file
    dx = 1500
    data = np.zeros((2 * dx + 1, 2 * dx + 1))
    # hdu container
    hdu = fits.PrimaryHDU(data)
    pixel = 0.01 * (u.arcsec).to(u.deg)
    #
    hdu.header["CTYPE1"] = "RA---TAN"
    hdu.header["CTYPE2"] = "DEC--TAN"
    hdu.header["CUNIT1"] = "deg"
    hdu.header["CUNIT2"] = "deg"
    hdu.header["CDELT1"] = -1 * pixel
    hdu.header["CDELT2"] = pixel
    hdu.header["CRVAL1"] = ra0
    hdu.header["CRVAL2"] = dec0
    hdu.header["CRPIX1"] = dx * 1.0
    hdu.header["CRPIX2"] = dx * 1.0
    hdu.header["RADESYS"] = "FK5"
    hdu.header["EQUINOX"] = 2000.0
    hdu.header["BUNIT"] = "km/s"
    hdu.header["BMAJ"] = 4 * pixel
    hdu.header["BMIN"] = 4 * pixel
    hdu.header["BPA"] = 0.0
    hdu.writeto(file_in, overwrite=True)

    return str(file_in)

def test_convolve_Vlsr(generate_dummy_file) -> None:
    """Test the convolve_Vlsr function."""

    # Load the dummy FITS file
    file_in = generate_dummy_file
    with fits.open(file_in) as hdul:
        header = hdul[0].header
        data = hdul[0].data

    # Create a dummy V_lsr map
    V_lsr = np.random.random(data.shape)

    # Test convolve_Vlsr without dilation
    convolved_data = velocity_tools.keplerian_field.convolve_Vlsr(
        V_lsr, header, dilation=False)
    assert convolved_data.shape == V_lsr.shape
    assert not np.isnan(convolved_data).all()

    # Test convolve_Vlsr with dilation
    convolved_data_dilated = velocity_tools.keplerian_field.convolve_Vlsr(
        V_lsr, header, dilation=True)
    assert convolved_data_dilated.shape == V_lsr.shape
    assert not np.isnan(convolved_data_dilated).all()

def test_generate_Vlsr(generate_dummy_file) -> None:
    # Load the dummy FITS file
    file_in = generate_dummy_file
    with fits.open(file_in) as hdul:
        header = hdul[0].header
        data = hdul[0].data
    ra0 = header['CRVAL1'] * u.deg
    dec0 = header['CRVAL2'] * u.deg
    R_out = 300.*u.au
    # Generate the Vlsr map
    inc0 = 42.*u.deg
    # inc2 = 60.*u.deg
    results = velocity_tools.keplerian_field.generate_Vlsr(
        header, ra0, dec0, PA_Angle=142.*u.deg, inclination=inc0, 
        distance=110.02*u.pc, R_out=R_out, Mstar=2.2*u.Msun, 
        Vc=5.2*u.km/u.s)
    # results2 = velocity_tools.keplerian_field.generate_Vlsr(
    #     header, ra0, dec0, PA_Angle=142.*u.deg, inclination=inc2, 
    #     distance=110.02*u.pc, R_out=R_out, Mstar=2.2*u.Msun, 
    #     Vc=5.2*u.km/u.s, do_plot=False)
    assert results is not None
    assert results.v.shape == data.shape
    # Check that values outside R_out are NaN
    assert np.isnan(results.v[results.r > R_out]).all()


# def test_generate_Vlsr() -> None:
#     rdisk = 300 * u.au
#     radius = np.linspace(1, 10, endpoint=True, num=5) * u.au
#     angle = 0 * u.deg
#     inclination = 0 * u.deg
#     vc0 = 5 * u.km / u.s

#     vel0 = velocity_tools.coordinate_offsets.generate_Vlsr(
#         radius,
#         angle,
#         inclination=inclination,
#         R_out=rdisk,
#         Mstar=1.0 * u.Msun,
#         Vc=0 * u.km / u.s,
#     )
#     vel1 = velocity_tools.coordinate_offsets.generate_Vlsr(
#         radius, 
#         angle, 
#         inclination=inclination, 
#         R_out=rdisk, 
#         Mstar=1.0 * u.Msun, 
#         Vc=vc0
#     )
#     assert ((vel1 - vel0) == vc0).all
